#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
Miscellany of wrapper scripts for command-line bioinformatics tools, public data downloaders and other generic routines.
"""

from jcvi.apps.base import dmain


if __name__ == '__main__':
    dmain(__file__)
