#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
Implementations of several key algorithms, such as: TSP, Graph, SuperMap, Linear Programming, ML, etc. used by other modules.
"""

from jcvi.apps.base import dmain


if __name__ == '__main__':
    dmain(__file__)
